@@ -1,41 +1,13 @@
-# the following two packages can be used to collect podparser
-# output into a string or simulate a portable null device
-
-package StrHandle;
-
-sub new {
-  my $type = shift;
-  my $this = bless {},$type;
-  $this->{Text} = "";
-  return $this;
-}
-
-sub print {
-  my $this = shift;
-  $this->{Text} .= "@_";
-}
-
-sub text {
-  $_[0]->{Text} .= $_[1] if $#_ > 0;
-  return $_[0]->{Text};
-}
-
-
-package NullHandle;
-
-@ISA = qw/ StrHandle/;
-
-sub print {} # do nothing
-
-
 # the filter for the PDL pod format (which is a valid general perl
 # pod format but with special interpretation of some =for directives)
 
 package PDL::PodParser;
 use PDL::Core '';
-use PDL::Pod::Parser;
+use Pod::Select;
+use File::Spec;
+use File::Basename;
 
-@ISA = qw(PDL::Pod::Parser);
+@ISA = qw(Pod::Select);
 
 %Title = ('Example' => 'Example',
 	  'Ref'     => 'Reference',
@@ -45,9 +17,13 @@ use PDL::Pod::Parser;
           'Bad'     => 'Bad value support',  
 	 );
 
+
+
 sub new {
-  my ($type) = @_;
-  my $parser = new PDL::Pod::Parser;
+  my $class = shift;
+  my $parser = $class->SUPER::new(@_);
+  bless $parser,$class; # just in case
+
   $parser->select("METHODS|OPERATORS|CONTRUCTORS|FUNCTIONS|NAME");
   $parser->{CURFUNC} = undef;
   $parser->{SYMHASH} = {};
@@ -55,12 +31,11 @@ sub new {
   $parser->{Mode} = "";
   $parser->{verbose} = 0;
   $parser->{NAME} = 'UNKNOWN';
-  bless $parser,$type;
   return $parser;
 }
 
 sub command {
-  my ($this,$cmd,$txt,$sep) = @_;
+  my ($this,$cmd,$txt,$line_num,$pod_para) = @_;
   $this->{Parmode} = 'Body';
 
   if ($cmd eq 'head1') {
@@ -68,7 +43,7 @@ sub command {
     $this->{Parmode} = 'Body';
     $this->{Parmode} = 'NAME' if $txt =~ /NAME/;
   } elsif ($this->{Mode} =~ /NAME/) {
-    last;
+    # do nothing (was 'last' but that was probably a mistake)
   } elsif ($cmd eq 'head2') {
     # A function can have multiple names (ex: zeros and zeroes),
     # so split at the commas
@@ -94,14 +69,14 @@ sub command {
     print "\n\tSignature: $sig\n" if defined $sig && $this->{verbose};
     $this->{CURFUNC} = $name;
   } elsif ($cmd eq 'for') {
-    $this->check_for_mode($txt,$sep) if $cmd eq 'for';
+    $this->check_for_mode($txt,$pod_para) if $cmd eq 'for';
   }
   local $this->{Parmode} = 'Body';
-  $this->SUPER::command($cmd,$txt,$sep);
+  $this->SUPER::command($cmd,$txt,$line_num,$pod_para);
 }
 
 sub check_for_mode {
-  my ($this,$txt,$sep) = @_;
+  my ($this,$txt,$pod_para) = @_;
   if ($txt =~ /^(sig|example|ref|opt|usage|bad|body)/i) {
     $this->{Parmode} = ucfirst lc $1;
     print "switched now to '$1' mode\n" if $this->{VERBOSE};
@@ -111,10 +86,11 @@ sub check_for_mode {
 }
 
 sub textblock {
-  my ($this,$txt) = @_;
+  my $this = shift;
+  my $txt = shift;
   $this->checkmode($txt);
   local $this->{INBLOCK} = 1;
-  $this->SUPER::textblock($txt);
+  $this->SUPER::textblock($txt,@_);
   $this->{Parmode} = 'Body'; # and reset parmode
 }
 
@@ -141,9 +117,10 @@ sub checkmode {
 }
 
 sub verbatim {
-  my ($this,$txt) = @_;
+  my $this = shift;
+  my $txt = shift;
   $this->checkmode($txt,1);
-  $this->SUPER::verbatim($txt);
+  $this->SUPER::verbatim($txt,@_);
 }
 
 # this needs improvement
@@ -185,6 +162,22 @@ PDL::Doc - support for PDL online documentation
 
 An implementation of online docs for PDL.
 
+=head1 Using PDL documentation
+
+PDL::Doc's main use is in the "help" (synonym "?") and "apropos"
+(synonym "??") commands in the perldl shell.  PDL:Doc provides the
+infrastrucure to index and access PDL's documentation through these
+commands.  There is also an API for direct access to the documentation 
+database (see below).
+
+The PDL doc system is built on Perl's pod (Plain Old Documentation),
+included inline with each module. The PDL core modules are
+automatically indexed when PDL is built and installed, and there is
+provision for indexing external modules as well.
+
+To include your module's pod into the Perl::Doc index, you should
+follow the documentation conventions below.  
+
 =head1 PDL documentation conventions
 
 For a package like PDL that has I<a lot> of functions it
@@ -215,18 +208,6 @@ Individual functions or methods in these section are introduced by
 
   =head2 funcname
 
-=cut
-
-# XXX - I don't think this is going to work for now -
-# it garbles Pod::Parser. I am changing all occurences
-# to use =for sig for now - KGB
-
-#or
-#  =head2 funcname(signature)
-
-
-=pod
-
 where signature is the argumentlist for a PP defined function as
 explained in L<PDL::PP>. Generally, PDL documentation is in valid POD
 format (see L<perlpod|perlpod>) but uses the C<=for> directive in a
@@ -234,13 +215,6 @@ special way. The C<=for> directive is used to flag to the PDL Pod
 parser that information is following that will be used to generate
 online help.
 
-The PDL podparser is derived from the PDL::Pod::Parser class that had
-to be patched in a few places, partly to fix minor bugs, partly
-to enhance functionality for perusal by PDL::Doc. Since the PDL::Doc
-module is still experimental the patched Pod-Parser distribution is
-included with the current PDL-Doc distribution. Note that PDL::Doc
-will I<not> work correctly with the released Pod-Parser distribution.
-
 The PDL Pod parser recognises the following C<=for> directives:
 
 =over 5
@@ -301,9 +275,10 @@ gives examples of typical usage for the current function:
 =item Bad
 
 provides information on how the function handles bad values (if
-C<$PDL:Config{WITH_BADVAL}> is set to 1). The intention is to
-have this information automatically created for pp-compiled
-functions, although it can be over-ridden.
+C<$PDL:Config{WITH_BADVAL}> is set to 1). The documentation under
+this directive should indicate if this function accepts piddles
+with bad values and under what circumstances this function might
+return piddles with bad values.
 
 =back
 
@@ -404,7 +379,7 @@ documented along these lines should have is the I<Ref> field preceding
 a one line description of its intended functionality (suitable for
 inclusion in a concise reference card). PP defined functions (see L<PDL::PP>)
 should have a I<Sig> field stating their signature. To facilitate
-maintainance of this documentation for such functions the 'Doc' field
+maintenance of this documentation for such functions the 'Doc' field
 has been introduced into the definition of C<pp_def> (see again L<PDL::PP>)
 which will take care that name and signature of the so defined function
 are documented in this way (for examples of this usage see, for example,
@@ -496,7 +471,11 @@ sub ensuredb {
     while (read IN, $plen,2) {
       my ($len) = unpack "S", $plen;
       read IN, $txt, $len;
-      my ($sym, %hash) = split chr(0), $txt;
+      my (@a) =  split chr(0), $txt;
+      push(@a, "") unless(@a % 2);  # Add null string at end if necessary -- solves bug with missing REF section.
+      my ($sym, %hash) = @a;
+     
+      $hash{Dbfile} = $fi; # keep the origin pdldoc.db path
       $this->{SYMS}->{$sym} = {%hash};
     }
     close IN;
@@ -518,6 +497,13 @@ sub savedb {
   open OUT, ">$this->{Outfile}" or barf "can't write to symdb $this->{Outfile}";
   binmode OUT;
   while (my ($key,$val) = each %$hash) {
+    next if 0 == scalar(%$val);
+    my $fi = $val->{File};
+    if (File::Spec->file_name_is_absolute($fi) && -f $fi) {
+      #store paths to *.pm files relative to pdldoc.db
+      $val->{File} = File::Spec->abs2rel($fi, dirname($this->{Outfile})) ;
+    }
+    delete $val->{Dbfile}; # no need to store Dbfile
     my $txt = "$key".chr(0).join(chr(0),%$val);
     print OUT pack("S",length($txt)).$txt;
   }
@@ -584,7 +570,7 @@ that should be matched against the regex. Valid fields are
   Ref,     # the one-line reference description
   Example, # the example for this function
   Opt,     # options
-  File,    # the path to the source file this docs have been extracted from
+  File,    # the path to the source file these docs have been extracted from
 
 If you wish to have your results sorted by function name, pass a true
 value for C<$sort>.
@@ -677,7 +663,7 @@ sub scan {
   my $infile =  new IO::File $file;
   # XXXX convert to absolute path
   # my $outfile = '/tmp/'.basename($file).'.pod';
-  my $outfile = new NullHandle;
+  open my $outfile, '>', \(my $outfile_text);
 
   # Handle RPM etc. case where we are building away from the final
   # location. Alright it's a hack - KGB
@@ -686,7 +672,9 @@ sub scan {
 
   my $parser = new PDL::PodParser;
   $parser->{verbose} = $verbose;
-  $parser->parse_from_filehandle($infile,$outfile);
+  eval { $parser->parse_from_filehandle($infile,$outfile) };
+  warn "cannot parse '$file'" if $@;
+
   $this->{SYMS} = {} unless defined $this->{SYMS};
   my $hash = $this->{SYMS};
   my @stats = stat $file;
@@ -707,15 +695,17 @@ sub scan {
   # and one can now find modules with 'apropos'
 
   $infile =  new IO::File $file;
-  $outfile = new StrHandle;
+  $outfile_text = '';
   $parser = new PDL::PodParser;
   $parser->select('NAME');
-  $parser->parse_from_filehandle($infile,$outfile);
-  my @namelines = split("\n",$outfile->{Text});
+  eval { $parser->parse_from_filehandle($infile,$outfile) };
+  warn "cannot parse '$file'" if $@;
+
+  my @namelines = split("\n",$outfile_text);
   my ($name,$does);
   for (@namelines) {
-     if (/^(PDL) (-) (.*)/ or /\s*(PDL::[\w:]*)\s*(-*)?\s*(.*)\s*$/) {
-       $name = $1; $does = $3;
+     if (/^(PDL) (-) (.*)/ or  /^\s*(Inline::Pdlpp)\s*(-*)?\s*(.*)\s*$/ or /\s*(PDL::[\w:]*)\s*(-*)?\s*(.*)\s*$/) {
+	$name = $1; $does = $3;
      }
      if (/^\s*([a-z][a-z0-9]*) (-+) (.*)/) { # lowercase shell script name
        $name = $1; $does = $3;
@@ -773,7 +763,12 @@ sub funcdocs {
   my ($this,$func,$fout) = @_;
   my $hash = $this->ensuredb;
   barf "unknown function '$func'" unless defined($hash->{$func});
-  funcdocs_fromfile($func,$hash->{$func}->{File},$fout);
+  my $file = $hash->{$func}->{File};
+  my $dbf = $hash->{$func}->{Dbfile};
+  if (!File::Spec->file_name_is_absolute($file) && $dbf) {
+    $file = File::Spec->rel2abs($file, dirname($dbf)); 
+  }
+  funcdocs_fromfile($func,$file,$fout);
 }
 
 =head1 FUNCTIONS
@@ -798,14 +793,14 @@ sub funcdocs_fromfile {
 
 sub extrdoc {
   my ($func,$file) = @_;
-  my $out = new StrHandle;
+  open my $out, '>', \(my $out_text);
   funcdocs_fromfile($func,$file,$out);
-  return $out->text;
+  return $out_text;
 }
 
 sub getfuncdocs {
   my ($func,$in,$out) = @_;
-  my $parser = new PDL::Pod::Parser;
+  my $parser = Pod::Select->new;
 #  $parser->select("\\(METHODS\\|OPERATORS\\|CONSTRUCTORS\\|FUNCTIONS\\|METHODS\\)/$func(\\(.*\\)*\\s*");
   foreach my $foo(qw/FUNCTIONS OPERATORS CONSTRUCTORS METHODS/) {
       seek $in,0,0;
@@ -814,6 +809,73 @@ sub getfuncdocs {
   }
 }
 
+
+=head2 add_module
+
+=for usage
+
+ use PDL::Doc; PDL::Doc::add_module("my::module");
+
+=for ref
+
+The C<add_module> function allows you to add POD from a particular Perl
+module that you've installed somewhere in @INC.  It searches for the
+active PDL document database and the module's .pod and .pm files, and
+scans and indexes the module into the database.
+
+C<add_module> is meant to be added to your module's Makefile as part of the
+installation script.
+
+=cut
+
+package PDL::Doc;
+sub add_module {
+    my($module) = shift;
+
+    use File::Copy qw{copy};
+
+    my($dir, $file, $pdldoc);
+    local($_);
+
+  DIRECTORY:
+    for(@INC){
+	$dir = $_;
+	$file = $dir."/PDL/pdldoc.db";
+	if( -f $file) {
+	    if(! -w "$dir/PDL") {
+		die "No write permission at $dir/PDL - not updating docs database.\n";
+	    }
+
+	    print "Found docs database $file\n";
+	    $pdldoc = new ("PDL::Doc",($file));
+	    last DIRECTORY;
+	}
+    }
+
+    die "Unable to find docs database - therefore not updating it.\n" unless($pdldoc);
+
+    my $mfile = $module;
+    $mfile =~ s/\:\:/\//g;
+    for(@INC){
+	my $postfix;
+	my $hit = 0;
+	for $postfix(".pm",".pod") {
+	    my $f = "$_/$mfile$postfix";
+	    if( -e $f ){
+		$pdldoc->ensuredb();
+		$pdldoc->scan($f);
+		eval { $pdldoc->savedb(); };
+		warn $@ if $@;
+		print "PDL docs database updated - added $f.\n";
+		$hit = 1;
+	    }
+	}
+	return if($hit);
+    }
+    
+    die "Unable to find a .pm or .pod file in \@INC for module $module\n";
+}
+
 1;
 
 =head1 PDL::DOC EXAMPLE
@@ -839,7 +901,7 @@ own code.
  print $pdldoc->gethash->{zeroes}->{Ref};
  
  # See which examples use zeroes
- $pdldoc->search('zeroes', 'Examples', 1);
+ $pdldoc->search('zeroes', 'Example', 1);
  
  # All the functions that use zeroes in their example:
  my @entries = $pdldoc->search('zeroes', 'Example', 1);
@@ -876,7 +938,7 @@ The Ref entry will begin with 'Module:' if it's a module. In code:
 =head1 BUGS
 
 Quite a few shortcomings which will hopefully be fixed following
-discussions on the pdl-porters mailing list.
+discussions on the pdl-devel mailing list.
 
 =head1 AUTHOR
 
